# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from contextlib import contextmanager
from typing import Any
from typing import Dict
from typing import Iterator
from typing import TypeVar

from kolena.errors import FrozenObjectError

T = TypeVar("T")


class Frozen:
    _frozen = False

    def _freeze(self) -> None:
        object.__setattr__(self, "_frozen", True)

    @contextmanager
    def _unfrozen(self) -> Iterator[None]:
        object.__setattr__(self, "_frozen", False)
        yield
        self._freeze()

    def _fields(self) -> Dict[str, Any]:
        return {key: value for key, value in self.__dict__.items() if not key.startswith("_")}

    def __setattr__(self, key: str, value: T) -> None:
        if self._frozen:
            raise FrozenObjectError("cannot set attribute of frozen object")
        object.__setattr__(self, key, value)

    def __repr__(self) -> str:
        fields = ", ".join(f"{key}={repr(value)}" for key, value in self._fields().items())
        return f"{type(self).__name__}({fields})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self.__dict__ == other.__dict__



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kolena._api.v1.detection import Metrics
from kolena.detection._internal import TestConfig as _TestConfig
from kolena.errors import InputValidationError


class TestConfig(_TestConfig):
    """
    Base class for testing configurations.

    See concrete implementations [`FixedGlobalThreshold`][kolena.detection.FixedGlobalThreshold] and
    [`F1Optimal`][kolena.detection.F1Optimal] for details.
    """


class FixedGlobalThreshold(TestConfig):
    """
    Test configuration that sets the default display threshold in Kolena to a fixed global threshold for all classes
    within the test suit.
    """

    fixed_threshold: float
    """
    The threshold used as the default for all classes when visualizing results in Kolena. Must be between 0 and 1.
    """

    iou_threshold: float
    """
    The minimum intersection over union (IoU) score between an inference and a ground truth for it to qualify as a
    potential match. Must be between 0 and 1.
    """

    def __init__(self, fixed_threshold: float, iou_threshold: float = 0.5):
        if iou_threshold < 0 or iou_threshold > 1:
            raise InputValidationError(f"iou_threshold of {iou_threshold} was not between 0 and 1")
        if fixed_threshold < 0 or fixed_threshold > 1:
            raise InputValidationError(f"threshold of {fixed_threshold} was not between 0 and 1")

        self.iou_threshold = iou_threshold
        self.threshold = fixed_threshold

    def _to_run_config(self) -> Metrics.RunConfig:
        return Metrics.RunConfig(
            strategy=Metrics.RunStrategy.FIXED_GLOBAL_THRESHOLD,
            iou_threshold=self.iou_threshold,
            params=dict(threshold=self.threshold),
        )


class F1Optimal(TestConfig):
    """
    Test configuration that sets the default display threshold in Kolena to the threshold that corresponds to the
    highest F1 score across all images within the test suite.

    This threshold is evaluated and set per class for test suites with multiple classes.
    """

    iou_threshold: float
    """
    The minimum intersection over union (IoU) score between an inference and a ground truth for it to qualify as a
    potential match. Must be between 0 and 1.
    """

    def __init__(self, iou_threshold: float = 0.5):
        if iou_threshold < 0 or iou_threshold > 1:
            raise InputValidationError(f"iou_threshold of {iou_threshold} was not between 0 and 1")
        self.iou_threshold = iou_threshold

    def _to_run_config(self) -> Metrics.RunConfig:
        return Metrics.RunConfig(strategy=Metrics.RunStrategy.F1_OPTIMAL, iou_threshold=self.iou_threshold)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Optional

from kolena._api.v1.core import TestCase as API
from kolena._api.v1.workflow import WorkflowType
from kolena.classification import TestCase
from kolena.detection._internal import BaseTestSuite


class TestSuite(BaseTestSuite):
    """
    A test suite is a grouping of [`TestCase`][kolena.classification.TestCase] objects.

    Testing on test suites is performed via [`test`][kolena.classification.test]. Metrics are computed across all
    samples in a test suite and also for each individual test case within the suite.

    For additional functionality, see the associated
    [base class documentation][kolena.detection._internal.test_suite.BaseTestSuite].

    :param name: The name of the test suite. If a test suite by this name already exists, that test suite is loaded.
    :param version: Optionally specify the version of the test suite to load. Ignored when the a suite by the
        provided name does not already exist.
    :param description: Optionally specify a description for the new test suite. Ignored when a test suite with the
        provided name already exists.
    :param test_cases: Optionally provide a list of [`TestCase`][kolena.classification.TestCase] tests used to seed a
        new test suite. Ignored when a test suite with the provided name already exists.
    """

    name: str
    """Unique name of the test suite."""

    version: int
    """
    The version of the test suite. Version is automatically incremented whenever the test suite is modified via
    [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
    """

    description: str
    """
    Free-form description of this test suite. May be edited at any time via
    [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
    """

    test_cases: List[TestCase]
    """
    The [`TestCase`][kolena.classification.TestCase] objects in this test suite. May be edited at any time via
    [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
    """

    _id: int
    _workflow: WorkflowType = WorkflowType.CLASSIFICATION

    def __init__(
        self,
        name: str,
        version: Optional[int] = None,
        description: Optional[str] = None,
        test_cases: Optional[List[TestCase]] = None,
        reset: bool = False,
    ):
        super().__init__(name, WorkflowType.CLASSIFICATION, version, description, test_cases, reset)

    @classmethod
    def _test_case_from(cls, response: API.EntityData) -> TestCase:
        return TestCase._from(response)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ._activation_map import colorize_activation_map
from ._activation_map import Colormap
from ._activation_map import ColormapJet
from ._utils import encode_png

__all__ = [
    "Colormap",
    "ColormapJet",
    "colorize_activation_map",
    "encode_png",
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from pydantic import Extra

from kolena._utils.consts import FieldName


class ValidatorConfig:
    """Pydantic configuration for dataclasses and @validate_arguments decorators."""

    arbitrary_types_allowed = True
    smart_union = True
    extra = Extra.allow  # do not fail when unrecognized values are provided


def validate_name(field: str, field_name: Optional[FieldName] = None):
    field_name_str = field_name.value if field_name else "field"
    if not field or field.isspace():
        raise ValueError(f"{field_name_str} must be non empty")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial

from kolena._utils import krequests
from kolena._utils.state import API_V2

get = partial(krequests.get, api_version=API_V2)
post = partial(krequests.post, api_version=API_V2)
put = partial(krequests.put, api_version=API_V2)
delete = partial(krequests.delete, api_version=API_V2)
raise_for_status = krequests.raise_for_status



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Optional

from pydantic import validate_arguments

from kolena._api.v1.workflow import WorkflowType
from kolena._utils.validators import ValidatorConfig
from kolena.detection import TestImage
from kolena.detection._datatypes import TestImageDataFrame
from kolena.detection._internal import BaseTestCase


class TestCase(BaseTestCase):
    """
    A test case holds a set of images to compute performance metrics against.

    For additional functionality, see the associated
    [base class documentation][kolena.detection._internal.test_case.BaseTestCase].

    :param name: Name of the test case to create or load.
    :param version: Optionally specify the version of the test case to load. When absent, the latest version is loaded.
        Ignored when creating new test cases.
    :param description: Optionally specify a description for a newly created test case. For existing test cases, this
        description can be edited via [`TestCase.edit`][kolena.detection._internal.test_case.BaseTestCase.edit].
    :param images: Optionally provide a list of images and associated ground truths to populate a new test case. For
        existing test cases, images can be edited via
        [`TestCase.edit`][kolena.detection._internal.test_case.BaseTestCase.edit].
    """

    _TestImageClass = TestImage
    _TestImageDataFrameClass = TestImageDataFrame

    _workflow = WorkflowType.DETECTION

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        name: str,
        version: Optional[int] = None,
        description: Optional[str] = None,
        images: Optional[List[_TestImageClass]] = None,
        reset: bool = False,
    ):
        super().__init__(name, WorkflowType.DETECTION, version, description, images, reset)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import Dict

from kolena._api.v1.detection import Metrics
from kolena._utils.frozen import Frozen
from kolena._utils.serializable import Serializable


class TestConfig(Frozen, Serializable, metaclass=ABCMeta):
    @abstractmethod
    def __init__(self) -> None:
        ...

    @abstractmethod
    def _to_run_config(self) -> Metrics.RunConfig:
        ...

    def _to_dict(self) -> Dict[str, Any]:
        run_config = self._to_run_config()
        return {k: v for k, v in run_config.__dict__.items() if not k.startswith("_")}
